/*
 * This file is open source software, licensed to you under the terms
 * of the Apache License, Version 2.0 (the "License").  See the NOTICE file
 * distributed with this work for additional information regarding copyright
 * ownership.  You may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
/*
 * Copyright 2023-present ScyllaDB
 */

#pragma once

#include <seastar/util/modules.hh>

/// \file

namespace seastar {

SEASTAR_MODULE_EXPORT_BEGIN

using shard_id = unsigned;

SEASTAR_MODULE_EXPORT_END

namespace internal {

#ifdef SEASTAR_BUILD_SHARED_LIBS
shard_id* this_shard_id_ptr() noexcept;
#else
inline shard_id* this_shard_id_ptr() noexcept {
    static thread_local shard_id g_this_shard_id;
    return &g_this_shard_id;
}
#endif

} // namespace internal

SEASTAR_MODULE_EXPORT_BEGIN

/// Returns shard_id of the of the current shard.
inline shard_id this_shard_id() noexcept {
    return *internal::this_shard_id_ptr();
}

SEASTAR_MODULE_EXPORT_END


} // namespace seastar
